<?php declare(strict_types=1);

namespace MyPrayerJournal;

use BitBadger\PDODocument\{DocumentException, DocumentList};
use DateTimeImmutable;
use DateTimeZone;
use Exception;

/**
 * User interface building blocks
 */
class UI
{
    /**
     * Generate a material icon
     *
     * @param string $name The name of the material icon
     * @return string The material icon wrapped in a `span` tag
     */
    public static function icon(string $name): string {
        return "<span class=material-icons>$name</span>";
    }

    /**
     * Render the journal items for the current user
     *
     * @throws DocumentException If any is encountered
     */
    public static function journal(): void
    {
        Layout::bareHead();
        $reqs = Request::forJournal();
        if ($reqs->hasItems()) { ?>
            <section id=journalItems class="row row-cols-1 row-cols-md-2 row-cols-lg-3 row-cols-xl-4 g-3" hx-target=this
                     hx-swap=outerHTML aria-label="Prayer Requests"><?php
            $spacer = '<span>&nbsp;</span>';
            foreach ($reqs->items() as /** @var Request $req */ $req) { ?>
                <div class=col>
                <div class="card h-100">
                    <div class="card-header p-0 d-flex" role=toolbar><?php
                        self::pageLink("/request/edit?id=$req->id", self::icon('edit'),
                            ['class' => 'btn btn-secondary', 'title' => 'Edit Request']); ?>
                        <?=$spacer?>
                        <button type=button class="btn btn-secondary" title="Add Notes" data-bs-toggle=modal
                                data-bs-target=#notesModal hx-get="/components/request/add-note?id=<?=$req->id?>"
                                hx-target=#notesBody hx-swap=innerHTML><?=self::icon('comment');?></button>
                        <?=$spacer?>
                        <button type=button class="btn btn-secondary" title="Snooze Request" data-bs-toggle=modal
                                data-bs-target=#snoozeModal hx-get="/components/request/snooze?id=<?=$req->id?>"
                                hx-target=#snoozeBody hx-swap=innerHTML><?=self::icon('schedule');?></button>
                        <div class=flex-grow-1></div>
                        <button type=button class="btn btn-success w-25" hx-patch="/request/prayed?id=<?=$req->id?>"
                                title="Mark as Prayed"><?=self::icon('done');?></button>
                    </div>
                    <div class=card-body>
                        <p class=request-text><?=htmlentities($req->currentText());?>
                    </div>
                    <div class="card-footer text-end text-muted px-1 py-0">
                        <em><?php
                            $lastPrayed = $req->lastPrayed();
                            echo 'last ' . (is_null($lastPrayed) ? 'activity': 'prayed') . ' ';
                            self::relativeDate($lastPrayed ?? $req->history[0]->asOf); ?>
                        </em>
                    </div>
                </div>
                </div><?php
            } ?>
            </section><?php
        } else {
            UI::noResults('No Active Requests', '/request/edit?id=new', 'Add a Request', <<<'TEXT'
                You have no requests to be shown; see the &ldquo;Active&rdquo; link above for snoozed or deferred
                requests, and the &ldquo;Answered&rdquo; link for answered requests
                TEXT);
        }
        Layout::bareFoot();
    }

    /**
     * Create a card when there are no results found
     */
    public static function noResults(string $heading, string $link, string $buttonText, string $text): void
    { ?>
        <div class=card>
        <h5 class=card-header><?=$heading?></h5>
        <div class="card-body text-center">
            <p class=card-text><?=$text?></p><?php
            self::pageLink($link, $buttonText, ['class' => 'btn btn-primary']); ?>
        </div>
        </div><?php
    }

    /**
     * Generate a link to a page within myPrayerJournal
     *
     * @param string $href The URL for the link
     * @param string $text The text for the link
     * @param array $attrs Any additional attributes that should be placed on the `a` tag
     */
    public static function pageLink(string $href, string $text, array $attrs = []): void
    { ?>
        <a href="<?=$href?>" hx-get="<?=$href?>" hx-target=#top hx-swap=innerHTML hx-push-url=true<?php
        foreach ($attrs as $key => $value) echo " $key=\"" . htmlspecialchars($value) . "\""; ?>><?=$text?></a><?php
    }

    /**
     * @throws Exception
     */
    public static function relativeDate(string $date): void
    {
        $parsed = new DateTimeImmutable($date);
        $inZone = $parsed->setTimezone(new DateTimeZone($_REQUEST['time_zone']));
        echo '<span title="' . date_format($inZone, 'l, F j, Y \a\t g:ia T') . '">'
            . self::formatDistance('now', $parsed) . '</span>';
    }

    //  Many thanks to date-fns (https://date-fns.org) for this logic
    /**
     * Format the distance between two dates
     *
     * @param string|DateTimeImmutable $from The starting date/time
     * @param string|DateTimeImmutable $to The ending date/time
     * @return string The distance between two dates
     * @throws Exception If date/time objects cannot be created
     */
    public static function formatDistance(string|DateTimeImmutable $from, string|DateTimeImmutable $to): string
    {
        $aDay        =  1_440.0;
        $almost2Days =  2_520.0;
        $aMonth      = 43_200.0;
        $twoMonths   = 86_400.0;

        $dtFrom  = is_string($from) ? new DateTimeImmutable($from) : $from;
        $dtTo    = is_string($to)   ? new DateTimeImmutable($to)   : $to;
        $minutes = abs($dtFrom->getTimestamp() - $dtTo->getTimestamp()) / 60;
        $months  = round($minutes / $aMonth);
        $years   = round($months / 12);

        $typeAndNumber = match (true) {
            $minutes < 1.0          => [FormatDistanceToken::LessThanXMinutes, 1],
            $minutes < 45.0         => [FormatDistanceToken::XMinutes,         round($minutes)],
            $minutes < 90.0         => [FormatDistanceToken::AboutXHours,      1],
            $minutes < $aDay        => [FormatDistanceToken::AboutXHours,      round($minutes / 60)],
            $minutes < $almost2Days => [FormatDistanceToken::XDays,            1],
            $minutes < $aMonth      => [FormatDistanceToken::XDays,            round($minutes / $aDay)],
            $minutes < $twoMonths   => [FormatDistanceToken::AboutXMonths,     round($minutes / $aMonth)],
            $months      < 12       => [FormatDistanceToken::XMonths,          round($minutes / $aMonth)],
            $months % 12 < 3        => [FormatDistanceToken::AboutXYears,      $years],
            $months % 12 < 9        => [FormatDistanceToken::OverXYears,       $years],
            default                 => [FormatDistanceToken::AlmostXYears,     $years]
        };
        $format = match ($typeAndNumber[0]) {
            FormatDistanceToken::LessThanXMinutes => ['less than a minute', 'less than %d minutes'],
            FormatDistanceToken::XMinutes         => ['a minute',           '%d minutes'],
            FormatDistanceToken::AboutXHours      => ['about an hour',      'about %d hours'],
            FormatDistanceToken::XHours           => ['an hour',            '%d hours'],
            FormatDistanceToken::XDays            => ['a day',              '%d days'],
            FormatDistanceToken::AboutXWeeks      => ['about a week',       'about %d weeks'],
            FormatDistanceToken::XWeeks           => ['a week',             '%d weeks'],
            FormatDistanceToken::AboutXMonths     => ['about a month',      'about %d months'],
            FormatDistanceToken::XMonths          => ['a month',            '%d months'],
            FormatDistanceToken::AboutXYears      => ['about a year',       'about %d years'],
            FormatDistanceToken::XYears           => ['a year',             '%d years'],
            FormatDistanceToken::OverXYears       => ['over a year',        'over %d years'],
            FormatDistanceToken::AlmostXYears     => ['almost a year',      'almost %d years']
        };
        $value = $typeAndNumber[1] == 1 ? $format[0] : sprintf($format[1], $typeAndNumber[1]);
        return $dtFrom > $dtTo ? "$value ago" : "in $value";
    }

    /**
     * Render the given list of requests
     *
     * @param DocumentList<Request> $reqs The list of requests to render
     * @throws Exception If date/time instances are not valid
     */
    public static function requestList(DocumentList $reqs): void
    {
        $btnClass   = "btn btn-light mx-2";
        $restoreBtn = fn(string $id, string $link, string $title) =>
            '<button class="' . $btnClass. '" hx-patch="/request/' . $link . '?id=' . $id
                . '" title="' . htmlspecialchars($title) . '">' . self::icon('restore') . '</button>'; ?>
        <div class=list-group><?php
            foreach ($reqs->items() as /** @var Request $req */ $req) { ?>
                <div class="list-group-item px-0 d-flex flex-row align-items-start" hx-target=this
                     hx-swap=outerHTML><?php
                    self::pageLink("/request/full?id=$req->id", self::icon('description'),
                        ['class' => $btnClass, 'title' => 'View Full Request']);
                    if (!$req->isAnswered()) {
                        self::pageLink("/request/edit?id=$req->id", self::icon('edit'),
                            ['class' => $btnClass, 'title' => 'Edit Request']);
                    }
                    if ($req->isSnoozed()) {
                        echo $restoreBtn($req->id, 'cancel-snooze', 'Cancel Snooze');
                    } elseif ($req->isPending()) {
                        echo $restoreBtn($req->id, 'show', 'Show Now');
                    }
                    echo '<p class="request-text mb-0">' . $req->currentText();
                    if ($req->isSnoozed() || $req->isPending() || $req->isAnswered()) { ?>
                        <br>
                        <small class=text-muted><em><?php
                            switch (true) {
                                case $req->isSnoozed():
                                    echo 'Snooze expires '; self::relativeDate($req->snoozedUntil);
                                    break;
                                case $req->isPending():
                                    echo 'Request appears next '; self::relativeDate($req->showAfter);
                                    break;
                                default:
                                    echo 'Answered '; self::relativeDate($req->history[0]->asOf);
                            } ?>
                        </em></small><?php
                    } ?>
                </div><?php
            } ?>
        </div><?php
    }
}

enum FormatDistanceToken
{
    case LessThanXMinutes;
    case XMinutes;
    case AboutXHours;
    case XHours;
    case XDays;
    case AboutXWeeks;
    case XWeeks;
    case AboutXMonths;
    case XMonths;
    case AboutXYears;
    case XYears;
    case OverXYears;
    case AlmostXYears;
}
